using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace PhSystems.Twain
{
	[StructLayout(LayoutKind.Sequential, Pack=2, CharSet=CharSet.Ansi)]
	internal struct TwCIEPOINT
	{
		TwFix32   X;
		TwFix32   Y;
		TwFix32   Z;
	} 

	[StructLayout(LayoutKind.Sequential, Pack=2, CharSet=CharSet.Ansi)]
	internal struct TwDECODEFUNCTION
	{
		TwFix32   StartIn;
		TwFix32   BreakIn;
		TwFix32   EndIn;
		TwFix32   StartOut;
		TwFix32   BreakOut;
		TwFix32   EndOut;
		TwFix32   Gamma;
		TwFix32   SampleCount;  /* if =0 use the gamma */
	} 

	[StructLayout(LayoutKind.Sequential, Pack=2, CharSet=CharSet.Ansi)]
	internal struct TwELEMENT8
	{
		char    Index;    /* Value used to index into the color table. */
		char    Channel1; /* First  tri-stimulus value (e.g Red)       */
		char    Channel2; /* Second tri-stimulus value (e.g Green)     */
		char    Channel3; /* Third  tri-stimulus value (e.g Blue)      */
	}

	[StructLayout(LayoutKind.Sequential, Pack=2, CharSet=CharSet.Ansi)]
	internal struct TwFRAME
	{
		TwFix32   Left;
		TwFix32   Top;
		TwFix32   Right;
		TwFix32   Bottom;
	} 


}


namespace PhSystems.Twain
{

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	internal struct TwFix32
	{												// TW_FIX32
		public short		Whole;
		public ushort		Frac;
	
		public float ToFloat()
		{
			return (float) Whole + ( (float)Frac /65536.0f );
		}
		public void FromFloat( float f )
		{
			int i = (int)((f * 65536.0f) + 0.5f);
			Whole = (short) (i >> 16);
			Frac = (ushort) (i & 0x0000ffff);
		}
	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	internal sealed  class TwUserInterface
	{									// TW_USERINTERFACE
		public short		ShowUI=0;				// bool is strictly 32 bit, so use short
		public short		ModalUI=0;
		public IntPtr		ParentHand=IntPtr.Zero;
		public void Initilalize(IntPtr hwnd)
		{
			this.ShowUI = (short)(hwnd==IntPtr.Zero?0:1);
			this.ModalUI = 1;
			this.ParentHand = hwnd;
		}
	}
		

	[StructLayout(LayoutKind.Sequential, Pack=2, CharSet=CharSet.Ansi)]
	public struct TwVersion
	{									// TW_VERSION
		public short		MajorNum;
		public short		MinorNum;
		public short		Language;
		public short		Country;
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst=34)]
		public string		Info;
	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public struct TwEvent
	{									// TW_EVENT
		public IntPtr		EventPtr;
		public short		Message;
	}

	[StructLayout(LayoutKind.Sequential, Pack=2, CharSet=CharSet.Ansi)]
	public sealed class TwIdentity
	{									// TW_IDENTITY
	
		public TwIdentity()
		{
			this.Version.Country=1;
			this.Version.Language=13;
			this.Version.MajorNum=1;
			this.Version.MinorNum=1;
		}

		public IntPtr		Id=IntPtr.Zero;
		public TwVersion	Version;
		public short		ProtocolMajor=0;
		public short		ProtocolMinor=0;
		public int			SupportedGroups=0;
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst=34)]
		public string		Manufacturer="";
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst=34)]
		public string		ProductFamily="";
		[MarshalAs(UnmanagedType.ByValTStr, SizeConst=34)]
		public string		ProductName="";
	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public sealed  class TwStatus
	{									// TW_STATUS
		public short		ConditionCode=0;		// TwCC
		public short		Reserved=0;
		public TwCC GetTwCC()
		{
			return (TwCC)this.ConditionCode;
		}
	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public sealed  class TwImageInfo
	{									// TW_IMAGEINFO
		public int			XResolution=0;
		public int			YResolution=0;
		public int			ImageWidth=0;
		public int			ImageLength=0;
		public short		SamplesPerPixel=0;
		[MarshalAs(UnmanagedType.ByValArray, SizeConst=8)] 
		public short[]		BitsPerSample=null;
		public short		BitsPerPixel=0;
		public short		Planar=0;
		public short		PixelType=0;
		public short		Compression=0;
	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public sealed  class TwPendingXfers
	{									// TW_PENDINGXFERS
		public short		Count=0;
		public int			EOJ=0;
	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public abstract class TwContainerCapability
	{ 
		public ushort ItemType = 0;

		public abstract	 TwOn GetTwOn();		

	}
	
	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public sealed class TwOneValue:TwContainerCapability
	{ 		
		public ulong Item = 0;

		public override	 TwOn GetTwOn()
		{
			return TwOn.One;
		}

		public static TwOneValue GetInstance(ushort ItemType,uint Item)
		{
			TwOneValue ret=new TwOneValue();
			ret.ItemType=ItemType;
			ret.Item=Item;
			return ret;
		}
	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public sealed class TwRange:TwContainerCapability
	{ 
		public uint MinValue = 0;
		public uint MaxValue = 0;
		public uint StepSize = 0;
		public uint DefaultValue = 0;
		public uint CurrentValue = 0;

		public override	 TwOn GetTwOn()
		{
			return TwOn.Range;
		}
	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public sealed class TwEnumeration:TwContainerCapability
	{ 
		public uint CurrentIndex = 0;
		public uint DefaultIndex = 0;
		public ushort[] ItemList=new ushort[]{}; 

		public override	 TwOn GetTwOn()
		{
			return TwOn.Enum;
		}


	}
	
	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public sealed class TwArray:TwContainerCapability
	{ 
		public ushort[] ItemList=new ushort[]{}; 
		
		public override	 TwOn GetTwOn()
		{
			return TwOn.Array;
		}

	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public sealed unsafe class TwCapability:IDisposable
	{									
		public short		Cap;
		public short		ConType;
		public IntPtr		Handle;

		~TwCapability()
		{
			this.Dispose();
		}

		public TwCapability( TwCap cap )
		{
			InitializeTwCapability(cap,TwOn.DontCare,null);
		}
		public TwCapability( TwCap cap, short sval)
		{
			InitializeTwCapability(cap,TwOn.One, TwOneValue.GetInstance((ushort)TwCapability.GetTwType(cap),(uint)sval));	
		}
		public TwCapability( TwCap cap, TwContainerCapability sval)
		{
			InitializeTwCapability(cap,sval.GetTwOn(),sval);	
		}

		
		private void InitializeTwCapability( TwCap cap,TwOn OnType, TwContainerCapability val)
		{
			Cap = (short) cap;
			ConType = (short) OnType;
			this.SetContainerCapability(val);
		}
						
		public void Dispose()
		{
			if( Handle != IntPtr.Zero )
				TwBaseManager.GlobalFree(Handle);
		}
		public void SetContainerCapability(TwContainerCapability cap)
		{
			if( Handle != IntPtr.Zero )
				TwBaseManager.GlobalFree(Handle);
			if (this.ConType!= (short)TwOn.DontCare || cap==null)
			{
				Handle = TwBaseManager.GlobalAlloc((int)(TwGMEM.MOVEABLE | TwGMEM.ZEROINIT),Marshal.SizeOf(cap.GetType()));
				Marshal.StructureToPtr(cap,TwSourceManager.GlobalLock(Handle),true);
				TwSourceManager.GlobalUnlock(Handle);
			}

		}
				
 		public TwContainerCapability GetContainerCapabilityValue()
		{
			TwContainerCapability ret=null;
			if (Handle!=IntPtr.Zero)
				if ((TwOn)this.ConType== TwOn.One) 
					ret=(TwContainerCapability)Marshal.PtrToStructure(TwSourceManager.GlobalLock(Handle), typeof(Twain.TwOneValue));
				if((TwOn)this.ConType== TwOn.Range)
					ret=(TwContainerCapability)Marshal.PtrToStructure(TwSourceManager.GlobalLock(Handle),  typeof(Twain.TwRange));
				if ((TwOn)this.ConType== TwOn.Array)
					ret=GetTwArray(TwSourceManager.GlobalLock(Handle));
				if ((TwOn)this.ConType== TwOn.Enum)
					ret=GetTwEnumeration(TwSourceManager.GlobalLock(Handle));
			if (ret!=null)
				TwSourceManager.GlobalUnlock(Handle);				
			return ret;
		}
					
		
		public static TwOn[] SupportedContainer(TwCap cap,bool TwMSG_GET)
		{
			if (cap==TwCap.AUDIOFILEFORMAT) 
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.XFERMECH) 
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.ALARMS) 
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.ALARMVOLUME) 
				return new TwOn[]{TwOn.Range,TwOn.One};
			if (cap==TwCap.AUTHOR) 
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.AUTOFEED) 
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.AUTOMATICCAPTURE) 
				return new TwOn[]{TwOn.Range,TwOn.One};
			if (cap==TwCap.AUTOSCAN) 
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.BATTERYMINUTES && TwMSG_GET) 
				return new TwOn[]{TwOn.Range,TwOn.One};
			if (cap==TwCap.BATTERYPERCENTAGE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.CAMERAPREVIEWUI && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.CAPTION && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.CAPTION && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.CLEARBUFFERS && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.CLEARBUFFERS && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.CLEARPAGE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.CLEARPAGE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.CUSTOMDSDATA && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.DEVICEEVENT && TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.DEVICEEVENT && !TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.DEVICEONLINE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.DEVICETIMEDATE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.DEVICETIMEDATE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.DUPLEX && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.DUPLEXENABLED && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.DUPLEXENABLED && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ENABLEDSUIONLY && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ENDORSER && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ENDORSER && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.EXTENDEDCAPS && TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.EXTENDEDCAPS && !TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.FEEDERALIGNMENT && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.FEEDERALIGNMENT && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.FEEDERENABLED && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.FEEDERENABLED && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.FEEDERLOADED && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.FEEDERORDER && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.FEEDERORDER && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.FEEDPAGE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.FEEDPAGE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.INDICATORS && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.INDICATORS && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.JOBCONTROL && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.JOBCONTROL && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.LANGUAGE && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.LANGUAGE && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.MAXBATCHBUFFERS && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.MAXBATCHBUFFERS && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.REACQUIREALLOWED && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.PAPERDETECTABLE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			//if (cap==TwCap.IPOWERSAVETIME && TwMSG_GET)
			//	return new TwOn[]{TwOn.One};
			//if (cap==TwCap.IPOWERSAVETIME && !TwMSG_GET)
			//	return new TwOn[]{TwOn.One};
			if (cap==TwCap.POWERSUPPLY && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.PRINTER && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.PRINTER && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.PRINTERENABLED && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.PRINTERENABLED && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.PRINTERINDEX && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.PRINTERINDEX && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.PRINTERMODE && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.PRINTERMODE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.PRINTERSTRING && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.PRINTERSTRING && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.PRINTERSUFFIX && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.PRINTERSUFFIX && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.REWINDPAGE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.REWINDPAGE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.SERIALNUMBER && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.SUPPORTEDCAPS && TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.TIMEBEFOREFIRSTCAPTURE && TwMSG_GET)
				return new TwOn[]{TwOn.One,TwOn.Range};
			if (cap==TwCap.TIMEBEFOREFIRSTCAPTURE && !TwMSG_GET)
				return new TwOn[]{TwOn.One,TwOn.Range};
			if (cap==TwCap.TIMEBETWEENCAPTURES && TwMSG_GET)
				return new TwOn[]{TwOn.One,TwOn.Range};
			if (cap==TwCap.TIMEBETWEENCAPTURES && !TwMSG_GET)
				return new TwOn[]{TwOn.One,TwOn.Range};
			if (cap==TwCap.TIMEDATE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.THUMBNAILSENABLED && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.THUMBNAILSENABLED && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.UICONTROLLABLE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.XFERCOUNT && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.XFERCOUNT && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IAUTOMATICBORDERDETECTION && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IAUTOMATICBORDERDETECTION && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IAUTOBRIGHT && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IAUTOBRIGHT && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IAUTOMATICDESKEW && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IAUTOMATICDESKEW && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IAUTOMATICROTATE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IAUTOMATICROTATE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IBARCODEDETECTIONENABLED && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IBARCODEDETECTIONENABLED && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IBARCODEMAXRETRIES && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBARCODEMAXRETRIES && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IBARCODEMAXSEARCHPRIORITIES && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBARCODEMAXSEARCHPRIORITIES && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IBARCODESEARCHMODE && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBARCODESEARCHMODE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IBARCODESEARCHPRIORITIES && TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.IBARCODESEARCHPRIORITIES && !TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.IBARCODETIMEOUT && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBARCODETIMEOUT && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IBITDEPTH && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBITDEPTH && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBITDEPTHREDUCTION && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBITDEPTHREDUCTION && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBITORDER && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBITORDER && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IBITORDERCODES && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBITORDERCODES && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IBRIGHTNESS && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IBRIGHTNESS && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ICCITTKFACTOR && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ICCITTKFACTOR && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ICOMPRESSION && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.ICOMPRESSION && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.ICONTRAST && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ICONTRAST && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ICUSTHALFTONE && TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.ICUSTHALFTONE && !TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.IEXPOSURETIME && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IEXPOSURETIME && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IEXTIMAGEINFO && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IEXTIMAGEINFO && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IFILTER && TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.IFILTER && !TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.IFLASHUSED && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IFLASHUSED && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IFLASHUSED2 && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.IFLASHUSED2 && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.IFLIPROTATION && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IFLIPROTATION && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IFRAMES && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IFRAMES && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IGAMMA && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IGAMMA && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IHALFTONES && TwMSG_GET)
				return new TwOn[]{TwOn.Array };
			if (cap==TwCap.IHALFTONES && !TwMSG_GET)
				return new TwOn[]{TwOn.Array };
			if (cap==TwCap.IHIGHLIGHT && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IHIGHLIGHT && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IIMAGEDATASET && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IIMAGEDATASET && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IIMAGEFILEFORMAT && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.IIMAGEFILEFORMAT && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.IIMAGEFILTER && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.IIMAGEFILTER && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.IJPEGPIXELTYPE && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IJPEGPIXELTYPE && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ILAMPSTATE && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ILAMPSTATE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ILIGHTPATH && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ILIGHTPATH && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ILIGHTSOURCE && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ILIGHTSOURCE && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IMAXFRAMES && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IMAXFRAMES && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IMINIMUMHEIGHT && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IMINIMUMWIDTH && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.INOISEFILTER && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.INOISEFILTER && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.IORIENTATION && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IORIENTATION && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IOVERSCAN && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.IOVERSCAN && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.IPATCHCODEDETECTIONENABLED && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPATCHCODEDETECTIONENABLED && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPATCHCODEMAXRETRIES && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IPATCHCODEMAXRETRIES && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPATCHCODEMAXSEARCHPRIORITIES && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IPATCHCODEMAXSEARCHPRIORITIES && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPATCHCODESEARCHMODE && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IPATCHCODESEARCHMODE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPATCHCODESEARCHPRIORITIES && TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.IPATCHCODESEARCHPRIORITIES && !TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.IPATCHCODETIMEOUT && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IPATCHCODETIMEOUT && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPHYSICALHEIGHT && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPHYSICALWIDTH && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPIXELFLAVOR && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IPIXELFLAVOR && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPIXELFLAVORCODES && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IPIXELFLAVORCODES && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IPIXELTYPE && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IPIXELTYPE && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IPLANARCHUNKY && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IPLANARCHUNKY && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IROTATION && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IROTATION && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ISHADOW && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ISHADOW && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ISUPPORTEDBARCODETYPES && TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.ISUPPORTEDPATCHCODETYPES && TwMSG_GET)
				return new TwOn[]{TwOn.Array};
			if (cap==TwCap.ISUPPORTEDSIZES && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.ISUPPORTEDSIZES && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One};
			if (cap==TwCap.ITHRESHOLD && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ITHRESHOLD && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.ITILES && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ITILES && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ITIMEFILL && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.ITIMEFILL && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IUNDEFINEDIMAGESIZE && TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IUNDEFINEDIMAGESIZE && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			if (cap==TwCap.IUNITS && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IUNITS && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.XFERMECH && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.XFERMECH && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IXNATIVERESOLUTION && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IXRESOLUTION && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IXRESOLUTION && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IXSCALING && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IXSCALING && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IYNATIVERESOLUTION && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IYRESOLUTION && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IYRESOLUTION && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IYSCALING && TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IYSCALING && !TwMSG_GET)
				return new TwOn[]{TwOn.Enum};
			if (cap==TwCap.IZOOMFACTOR && TwMSG_GET)
				return new TwOn[]{TwOn.Enum,TwOn.One,TwOn.Range};
			if (cap==TwCap.IZOOMFACTOR && !TwMSG_GET)
				return new TwOn[]{TwOn.One};
			return null;
		}

		public static System.Type GetEnumValuesCapability(TwCap cap)
		{
			if (cap==TwCap.IPIXELTYPE)
				return typeof(TWPT);
			if (cap==TwCap.IBITORDER)
				return typeof(TWBO); 
			if (cap==TwCap.ICOMPRESSION) 
				return typeof(TWCP); 
			if (cap==TwCap.IIMAGEFILEFORMAT)
				return typeof(TWFF);
			if (cap==TwCap.IFILTER)
				return typeof(TWFT); 
			if (cap==TwCap.ILIGHTPATH)
				return typeof(TWLP); 
			if (cap==TwCap.ILIGHTSOURCE)
				return typeof(TWLS); 
			if (cap==TwCap.IORIENTATION)
				return typeof(TWOR); 
			if (cap==TwCap.IPLANARCHUNKY)  
				return typeof(TWPC); 
			if (cap==TwCap.IPIXELFLAVOR)
				return typeof(TWPF); 
			if (cap==TwCap.IPIXELTYPE)   
				return typeof(TWPT); 
			if (cap==TwCap.ISUPPORTEDSIZES)   
				return typeof(TWSS); 
			if (cap==TwCap.XFERMECH)   
				return typeof(TWSX); 
			if (cap==TwCap.IUNITS)   
				return typeof(TWUN); 
			if (cap==TwCap.IBITDEPTHREDUCTION)   
				return typeof(TWBR); 
			if (cap==TwCap.JOBCONTROL)  
				return typeof(TWJC);
			if (cap==TwCap.AUDIOFILEFORMAT)  
				return typeof(TWAF); 
			if (cap==TwCap.ALARMS)  
				return typeof(TWAL); 
			if (cap==TwCap.CLEARBUFFERS)  
				return typeof(TWCB); 
			if (cap==TwCap.DEVICEEVENT)  
				return typeof(TWDE); 
			if (cap==TwCap.FEEDERALIGNMENT) 
				return typeof(TWFA); 
			if (cap==TwCap.FEEDERORDER)  
				return typeof(TWFO); 
			if (cap==TwCap.FILESYSTEM)  
				return typeof(TWFS); 
			if (cap==TwCap.POWERSUPPLY)  
				return typeof(TWPS); 
			if (cap==TwCap.PRINTER)  
				return typeof(TWPR); 
			if (cap==TwCap.PRINTERMODE)   
				return typeof(TWPM); 
			if (cap==TwCap.IBARCODESEARCHMODE)   
				return typeof(TWBD);
			if (cap==TwCap.IFLASHUSED2)   
				return typeof(TWFL); 
			if (cap==TwCap.IIMAGEFILTER)   
				return typeof(TWIF); 
			if (cap==TwCap.IOVERSCAN)  
				return typeof(TWOV);
			//if (cap==TwCap.DUPLEX)   
			//	return typeof(TWDX);
			//if (cap==TwCap.BARCODETYPE)   
			//	return typeof(TWBT);
			//if (cap==TwCap.DESKEWSTATUS)   
			//	return typeof(TWDSK);
			//if (cap==TwCap.PATCHCODE)   
			//	return typeof(TWPCH);
			//if (cap==TwCap.BARCODEROTATION) 
			//	return typeof(TWBCOR); 
			//if (cap==TwCap.FLIPROTATION)  
			//	return typeof(TWFR); 
			//if (cap==TwCap.NOISEFILTER)   
			//	return typeof(TWNF); 
			//if (cap==TwDAT.FileSystem)
			//	return typeof(TWFY); 
			return null;
		}	
		
		public static System.Type GetContainerCapabilityType(TwOn tipe)
		{
			if (tipe==TwOn.Array)
			{
				return  typeof(TwArray);
			}
			if (tipe==TwOn.Enum)
			{
				return  typeof(TwEnumeration);
			}	
			if (tipe==TwOn.One)
			{
				return   typeof(TwOneValue);
			}	
			if (tipe==TwOn.Range)
			{
				return   typeof(TwRange);
			}	
			return null;

		}


		internal static System.Type GetItemType(TwType ItemType)
		{
			if (ItemType==PhSystems.Twain.TwType.Frame)
				return typeof(Twain.TwFRAME);
			if (ItemType==PhSystems.Twain.TwType.Fix32)
				return typeof(Twain.TwFix32);
			if (ItemType==PhSystems.Twain.TwType.Bool)
				return typeof(ushort);
			if (ItemType==PhSystems.Twain.TwType.Int32)
				return typeof(long);
			if (ItemType==PhSystems.Twain.TwType.Int16)
				return typeof(short);
			if (ItemType==PhSystems.Twain.TwType.Int8)
				return typeof(char);
			if (ItemType==PhSystems.Twain.TwType.UInt32)
				return typeof(ulong);
			if (ItemType==PhSystems.Twain.TwType.UInt16)
				return typeof(ushort);
			if (ItemType==PhSystems.Twain.TwType.UInt8)
				return typeof(char);
			return typeof(ushort);
		}

		internal static TwType GetTwType(TwCap cap)
		{

			if (cap==TwCap.BATTERYPERCENTAGE)
				return TwType.Int16;
			if (cap==TwCap.CAMERAPREVIEWUI)
				return TwType.Bool;
			if (cap==TwCap.CAPTION)
				return TwType.Str255;
			if (cap==TwCap.CLEARBUFFERS)
				return TwType.UInt16;
			if (cap==TwCap.CLEARPAGE)
				return TwType.Bool;
			if (cap==TwCap.CUSTOMDSDATA)
				return TwType.Bool;
			if (cap==TwCap.DEVICEEVENT)
				return TwType.UInt16;
			if (cap==TwCap.DEVICEONLINE)
				return TwType.Bool;
			if (cap==TwCap.DEVICETIMEDATE)
				return TwType.Str32;
			if (cap==TwCap.DUPLEX)
				return TwType.UInt16;
			if (cap==TwCap.DUPLEXENABLED)
				return TwType.Bool;
			if (cap==TwCap.ENABLEDSUIONLY)
				return TwType.Bool;
			if (cap==TwCap.ENDORSER)
				return TwType.UInt32;
			if (cap==TwCap.EXTENDEDCAPS)
				return TwType.UInt16;
			if (cap==TwCap.FEEDERALIGNMENT)
				return TwType.UInt16;
			if (cap==TwCap.FEEDERENABLED)
				return TwType.Bool;
			if (cap==TwCap.FEEDERLOADED)
				return TwType.Bool;
			if (cap==TwCap.FEEDERORDER)
				return TwType.UInt16;
			if (cap==TwCap.FEEDPAGE)
				return TwType.Bool;
			if (cap==TwCap.INDICATORS)
				return TwType.Bool;
			if (cap==TwCap.JOBCONTROL)
				return TwType.UInt16;
			if (cap==TwCap.LANGUAGE)
				return TwType.UInt16;
			if (cap==TwCap.MAXBATCHBUFFERS)
				return TwType.UInt32;
			if (cap==TwCap.REACQUIREALLOWED)
				return TwType.Bool;
			if (cap==TwCap.PAPERDETECTABLE)
				return TwType.Bool;
			//if (cap==TwCap.IPOWERSAVETIME)
			//	return TwType.Int32;
			if (cap==TwCap.POWERSUPPLY)
				return TwType.UInt16;
			if (cap==TwCap.PRINTER)
				return TwType.UInt16;
			if (cap==TwCap.PRINTERENABLED)
				return TwType.Bool;
			if (cap==TwCap.PRINTERINDEX)
				return TwType.UInt32;
			if (cap==TwCap.PRINTERMODE)
				return TwType.UInt16;
			if (cap==TwCap.PRINTERSTRING)
				return TwType.Str255;
			if (cap==TwCap.PRINTERSUFFIX)
				return TwType.Str255;
			if (cap==TwCap.REWINDPAGE)
				return TwType.Bool;
			if (cap==TwCap.SERIALNUMBER)
				return TwType.Str255;
			if (cap==TwCap.SUPPORTEDCAPS)
				return TwType.UInt16;
			if (cap==TwCap.TIMEBEFOREFIRSTCAPTURE)
				return TwType.Int32;
			if (cap==TwCap.TIMEBETWEENCAPTURES)
				return TwType.Int32;
			if (cap==TwCap.TIMEDATE)
				return TwType.Str32;
			if (cap==TwCap.THUMBNAILSENABLED)
				return TwType.Bool;
			if (cap==TwCap.UICONTROLLABLE)
				return TwType.Bool;
			if (cap==TwCap.XFERCOUNT)
				return TwType.Int16;
			if (cap==TwCap.IAUTOMATICBORDERDETECTION)
				return TwType.Bool;
			if (cap==TwCap.IAUTOBRIGHT)
				return TwType.Bool;
			if (cap==TwCap.IAUTOMATICDESKEW)
				return TwType.Bool;
			if (cap==TwCap.IAUTOMATICROTATE)
				return TwType.Bool;
			if (cap==TwCap.IBARCODEDETECTIONENABLED)
				return TwType.Bool;
			if (cap==TwCap.IBARCODEMAXRETRIES)
				return TwType.UInt32;
			if (cap==TwCap.IBARCODEMAXSEARCHPRIORITIES)
				return TwType.UInt32;
			if (cap==TwCap.IBARCODESEARCHMODE)
				return TwType.UInt16;
			if (cap==TwCap.IBARCODESEARCHPRIORITIES)
				return TwType.UInt16;
			if (cap==TwCap.IBARCODETIMEOUT)
				return TwType.UInt32;
			if (cap==TwCap.IBITDEPTH)
				return TwType.UInt16;
			if (cap==TwCap.IBITDEPTHREDUCTION)
				return TwType.UInt16;
			if (cap==TwCap.IBITORDER)
				return TwType.UInt16;
			if (cap==TwCap.IBITORDERCODES)
				return TwType.UInt16;
			if (cap==TwCap.IBRIGHTNESS)
				return TwType.Fix32;
			if (cap==TwCap.ICCITTKFACTOR)
				return TwType.UInt16;
			if (cap==TwCap.ICOMPRESSION)
				return TwType.UInt16;
			if (cap==TwCap.ICONTRAST)
				return TwType.Fix32;
			if (cap==TwCap.ICUSTHALFTONE)
				return TwType.UInt8;
			if (cap==TwCap.IEXPOSURETIME)
				return TwType.Fix32;
			if (cap==TwCap.IEXTIMAGEINFO)
				return TwType.Bool;
			if (cap==TwCap.IFILTER)
				return TwType.UInt16;
			if (cap==TwCap.IFLASHUSED)
				return TwType.Bool;
			if (cap==TwCap.IFLASHUSED2)
				return TwType.UInt16;
			if (cap==TwCap.IFLIPROTATION)
				return TwType.UInt16;
			//if (cap==TwCap.IFRAMES)
			//	return TwType.FRAME;
			if (cap==TwCap.IGAMMA)
				return TwType.Fix32;
			if (cap==TwCap.IHALFTONES)
				return TwType.Str32;
			if (cap==TwCap.IHIGHLIGHT)
				return TwType.Fix32;
			if (cap==TwCap.IIMAGEDATASET)
				return TwType.UInt32;
			if (cap==TwCap.IIMAGEFILEFORMAT)
				return TwType.UInt16;
			if (cap==TwCap.IIMAGEFILTER)
				return TwType.UInt16;
			if (cap==TwCap.IJPEGPIXELTYPE)
				return TwType.UInt16;
			if (cap==TwCap.ILAMPSTATE)
				return TwType.Bool;
			if (cap==TwCap.ILIGHTPATH)
				return TwType.UInt16;
			if (cap==TwCap.ILIGHTSOURCE)
				return TwType.UInt16;
			if (cap==TwCap.IMAXFRAMES)
				return TwType.UInt16;
			if (cap==TwCap.IMINIMUMHEIGHT)
				return TwType.Fix32;
			if (cap==TwCap.IMINIMUMWIDTH)
				return TwType.Fix32;
			if (cap==TwCap.INOISEFILTER)
				return TwType.UInt16;
			if (cap==TwCap.IORIENTATION)
				return TwType.UInt16;
			if (cap==TwCap.IOVERSCAN)
				return TwType.UInt16;
			if (cap==TwCap.IPATCHCODEDETECTIONENABLED)
				return TwType.Bool;
			if (cap==TwCap.IPATCHCODEMAXRETRIES)
				return TwType.UInt32;
			if (cap==TwCap.IPATCHCODEMAXSEARCHPRIORITIES)
				return TwType.UInt32;
			if (cap==TwCap.IPATCHCODESEARCHMODE)
				return TwType.UInt16;
			if (cap==TwCap.IPATCHCODESEARCHPRIORITIES)
				return TwType.UInt16;
			if (cap==TwCap.IPATCHCODETIMEOUT)
				return TwType.UInt32;
			if (cap==TwCap.IPHYSICALHEIGHT)
				return TwType.Fix32;
			if (cap==TwCap.IPHYSICALWIDTH)
				return TwType.Fix32;
			if (cap==TwCap.IPIXELFLAVOR)
				return TwType.UInt16;
			if (cap==TwCap.IPIXELFLAVORCODES)
				return TwType.UInt16;
			if (cap==TwCap.IPIXELTYPE)
				return TwType.UInt16;
			if (cap==TwCap.IPLANARCHUNKY)
				return TwType.UInt16;
			if (cap==TwCap.IROTATION)
				return TwType.Fix32;
			if (cap==TwCap.ISHADOW)
				return TwType.Fix32;
			if (cap==TwCap.ISUPPORTEDBARCODETYPES)
				return TwType.UInt16;
			if (cap==TwCap.ISUPPORTEDPATCHCODETYPES)
				return TwType.UInt16;
			if (cap==TwCap.ISUPPORTEDSIZES)
				return TwType.UInt16;
			if (cap==TwCap.ITHRESHOLD)
				return TwType.Fix32;
			if (cap==TwCap.ITILES)
				return TwType.Bool;
			if (cap==TwCap.ITIMEFILL)
				return TwType.UInt16;
			if (cap==TwCap.IUNDEFINEDIMAGESIZE)
				return TwType.Bool;
			if (cap==TwCap.IUNITS)
				return TwType.UInt16;
			if (cap==TwCap.XFERMECH)
				return TwType.UInt16;
			if (cap==TwCap.IXNATIVERESOLUTION)
				return TwType.Fix32;
			if (cap==TwCap.IXRESOLUTION)
				return TwType.Fix32;
			if (cap==TwCap.IXSCALING)
				return TwType.Fix32;
			if (cap==TwCap.IYNATIVERESOLUTION)
				return TwType.Fix32;
			if (cap==TwCap.IYRESOLUTION)
				return TwType.Fix32;
			if (cap==TwCap.IYSCALING)
				return TwType.Fix32;
			if (cap==TwCap.IZOOMFACTOR)
				return TwType.Int16;
			return TwType.Int16;
		}




		private static TwArray GetTwArray(IntPtr pv)
		{
			TwArray result= new TwArray();
			unsafe
			{
				ushort* sp=(ushort*)pv.ToPointer();
				uint* ip;
				uint NumItems;
				result.ItemType=*sp;
				sp++;
				ip=(uint*)sp;
				NumItems=*ip;
				ip++;
				sp=(ushort*)ip;
				ushort[] llista=new ushort[NumItems];
				for (int i=0;i<NumItems;i++) 
				{
					llista[i]=*sp;
					sp++;
				}
				result.ItemList=llista;
			}
			return result;
		} 

		private static TwEnumeration GetTwEnumeration(IntPtr pv)
		{
			TwEnumeration result= new TwEnumeration();
			unsafe
			{
				ushort* sp=(ushort*)pv.ToPointer();
				uint* ip;
				uint NumItems;
				result.ItemType=*sp;
				sp++;
				ip=(uint*)sp;
				NumItems=*ip;
				ip++;
				result.CurrentIndex=*ip;
				ip++;
				result.DefaultIndex=*ip;
				ip++;
				sp=(ushort*)ip;
				ushort[] llista=new ushort[NumItems];
				for (int i=0;i<NumItems;i++) 
				{
					llista[i]=*sp;
					sp++;
				}
				result.ItemList=llista;
			}
			return result;
		} 


	}

	[StructLayout(LayoutKind.Sequential, Pack=2)]
	public sealed class TwBitmapInfoHeader
	{
		public int      biSize;
		public int      biWidth;
		public int      biHeight;
		public short    biPlanes;
		public short    biBitCount;
		public int      biCompression;
		public int      biSizeImage;
		public int      biXPelsPerMeter;
		public int      biYPelsPerMeter;
		public int      biClrUsed;
		public int      biClrImportant;
	}


	public class TwTriplet
	{
		internal TwDG DG;
		internal TwDAT DAT;
		internal TwMSG MSG;
		
		internal TwTriplet(TwDG DG,TwDAT DAT,TwMSG MSG)
		{
			this.DG=DG;
			this.DAT=DAT;
			this.MSG=MSG;
		}

		internal static TwOperation Operation(TwDG DG,TwDAT DAT,TwMSG MSG)
		{
			//if (TwDG.Audio   ==DG && TwDAT.AudioFILEXFER   == DAT && TwMSG.Get== MSG)  
			//	return TwOperation.AudioFileXferGet;
			//if (TwDG.Audio   ==DG && TwDAT.AudioINFO 		 == DAT && TwMSG.Get== MSG)  
			//	return TwOperation.AudioInfoGet;  
			//if (TwDG.Audio   ==DG && TwDAT.AudioNATIVEXFER == DAT && TwMSG.Get == MSG)  
			//	return TwOperation.AudioNativeXferGet;   
			if (TwDG.Control ==DG && TwDAT.Capability  	 == DAT && TwMSG.Get    == MSG)  
				return TwOperation.CapabilityGet;       
			if (TwDG.Control ==DG && TwDAT.Capability 	 == DAT && TwMSG.GetCurrent  == MSG)  
				return TwOperation.CapabilityGetCurrent;   
			if (TwDG.Control ==DG && TwDAT.Capability 	 == DAT && TwMSG.GetDefault  == MSG)  
				return TwOperation.CapabilityGetDefault;    
			if (TwDG.Control ==DG && TwDAT.Capability  	 == DAT && TwMSG.QuerySupport == MSG)  
				return TwOperation.CapabilityQuerySupport;   
			if (TwDG.Control ==DG && TwDAT.Capability 	 == DAT && TwMSG.Reset      == MSG)  
				return TwOperation.CapabilityReset;     
			if (TwDG.Control ==DG && TwDAT.Capability 	 == DAT && TwMSG.Set    == MSG)  
				return TwOperation.CapabilitySet;           
			if (TwDG.Image   ==DG && TwDAT.CieColor 		 == DAT && TwMSG.Get == MSG)  
				return TwOperation.CieColorGet;        
			if (TwDG.Control ==DG && TwDAT.CustomDSData 	 == DAT && TwMSG.Get == MSG)  
				return TwOperation.CustomDSDataGet;     
			if (TwDG.Control ==DG && TwDAT.CustomDSData    == DAT && TwMSG.Set       == MSG)  
				return TwOperation.CustomDSDataSet; 
			if (TwDG.Control ==DG && TwDAT.DeviceEvent 	 == DAT && TwMSG.Get        == MSG)  
				return TwOperation.DeviceEventGet; 
			if (TwDG.Control ==DG && TwDAT.Event 			 == DAT && TwMSG.ProcessEvent  == MSG)  
				return TwOperation.EventProcessEvent; 
			if (TwDG.Image   ==DG && TwDAT.ExtImageInfo 	 == DAT && TwMSG.Get          == MSG)  
				return TwOperation.ExtImageInfoGet;     
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.AutoCaptureDir == MSG)  
				return TwOperation.FileSystemAutomaticCaptureDirectory; 
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.ChangeDirectory      == MSG)  
				return TwOperation.FileSystemChangeDirectory; 
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.Copy              == MSG)  
				return TwOperation.FileSystemCopy; 	
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.CreateDirectory   == MSG)  
				return TwOperation.FileSystemCreateDirectory;
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.Delete      == MSG)  
				return TwOperation.FileSystemDelete; 
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.FormatMedia == MSG)  
				return TwOperation.FileSystemFormatMedia; 
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.GetClose    == MSG)  
				return TwOperation.FileSystemGetClose; 
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.GetFirstFile== MSG)  
				return TwOperation.FileSystemGetFirstFile; 
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.GetInfo     == MSG)  
				return TwOperation.FileSystemGetInfo; 
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.GetNextFile == MSG)  
				return TwOperation.FileSystemGetNextFile; 
			if (TwDG.Control ==DG && TwDAT.FileSystem 	 == DAT && TwMSG.Rename      == MSG)  
				return TwOperation.FileSystemRename; 
			if (TwDG.Image   ==DG && TwDAT.GrayResponse 	 == DAT && TwMSG.Reset == MSG)  
				return TwOperation.GrayscaleResponseReset;
			if (TwDG.Image   ==DG && TwDAT.GrayResponse 	 == DAT && TwMSG.Set   == MSG)  
				return TwOperation.GrayscaleResponseSet; 
			if (TwDG.Control ==DG && TwDAT.Identity 		 == DAT && TwMSG.CloseDS== MSG)  
				return TwOperation.IdentityCloseDS; 
			if (TwDG.Control ==DG && TwDAT.Identity 		 == DAT && TwMSG.GetDefault== MSG)  
				return TwOperation.IdentityGetDefault;
			if (TwDG.Control ==DG && TwDAT.Identity 		 == DAT && TwMSG.GetFirst  == MSG)  
				return TwOperation.IdentityGetFirst; 
			if (TwDG.Control ==DG && TwDAT.Identity 		 == DAT && TwMSG.GetNext   == MSG)  
				return TwOperation.IdentityGetNext; 
			if (TwDG.Control ==DG && TwDAT.Identity 		 == DAT && TwMSG.OpenDS    == MSG)  
				return TwOperation.IdentityOpenDS; 
			if (TwDG.Control ==DG && TwDAT.Identity 		 == DAT && TwMSG.UserSelect== MSG)  
				return TwOperation.IdentityUserSelect;  
			if (TwDG.Image   ==DG && TwDAT.ImageFileXfer 	 == DAT && TwMSG.Get       == MSG)  
				return TwOperation.ImageFileXferGet; 
			if (TwDG.Image   ==DG && TwDAT.ImageInfo 		 == DAT && TwMSG.Get       == MSG)  
				return TwOperation.ImageInfoGet;  
			if (TwDG.Image   ==DG && TwDAT.ImageLayout 	 == DAT && TwMSG.Get               == MSG)  
				return TwOperation.ImageLayoutGet;  
			if (TwDG.Image   ==DG && TwDAT.ImageLayout 	 == DAT && TwMSG.GetDefault        == MSG)  
				return TwOperation.ImageLayoutGetDefault; 
			if (TwDG.Image   ==DG && TwDAT.ImageLayout 	 == DAT && TwMSG.Reset             == MSG)  
				return TwOperation.ImageLayoutReset; 
			if (TwDG.Image   ==DG && TwDAT.ImageLayout 	 == DAT && TwMSG.Set               == MSG)  
				return TwOperation.ImageLayoutSet; 
			if (TwDG.Image   ==DG && TwDAT.ImageMemXfer 	 == DAT && TwMSG.Get       == MSG)  
				return TwOperation.ImageMemXferGet; 
			if (TwDG.Image   ==DG && TwDAT.ImageNativeXfer == DAT && TwMSG.Get             == MSG)  
				return TwOperation.ImageNativeXferGet;
			if (TwDG.Image   ==DG && TwDAT.JpegCompression == DAT && TwMSG.Get             == MSG)  
				return TwOperation.JpegCompressionGet; 
			if (TwDG.Image   ==DG && TwDAT.JpegCompression == DAT && TwMSG.GetDefault      == MSG)  
				return TwOperation.JpegCompressionGetDefault;
			if (TwDG.Image   ==DG && TwDAT.JpegCompression == DAT && TwMSG.Reset           == MSG)  
				return TwOperation.JpegCompressionReset; 
			if (TwDG.Image   ==DG && TwDAT.JpegCompression == DAT && TwMSG.Set             == MSG)  
				return TwOperation.JpegCompressionSet; 
			if (TwDG.Control ==DG && TwDAT.Null 			 == DAT && TwMSG.CloseDSReq== MSG)  
				return TwOperation.NullCloseDSReq; 
			if (TwDG.Control ==DG && TwDAT.Null  			 == DAT && TwMSG.DeviceEvent== MSG)  
				return TwOperation.NullDeviceEvent;
			if (TwDG.Control ==DG && TwDAT.Null 			 == DAT && TwMSG.XFerReady  == MSG)  
				return TwOperation.NullXferReady;
			if (TwDG.Image   ==DG && TwDAT.Palette8 		 == DAT && TwMSG.Get                == MSG)  
				return TwOperation.Palette8Get;  
			if (TwDG.Image   ==DG && TwDAT.Palette8 		 == DAT && TwMSG.GetDefault         == MSG)  
				return TwOperation.Palette8GetDefault;  
			if (TwDG.Image   ==DG && TwDAT.Palette8 		 == DAT && TwMSG.Reset              == MSG)  
				return TwOperation.Palette8Reset;  
			if (TwDG.Image   ==DG && TwDAT.Palette8 		 == DAT && TwMSG.Set                == MSG)  
				return TwOperation.Palette8Set; 
			if (TwDG.Control ==DG && TwDAT.PassThru 		 == DAT && TwMSG.PassThru           == MSG)  
				return TwOperation.PassThru; 	
			if (TwDG.Control ==DG && TwDAT.PendingXfers 	 == DAT && TwMSG.EndXfer            == MSG)  
				return TwOperation.PendingXfersEndXfer; 	
			if (TwDG.Control ==DG && TwDAT.PendingXfers 	 == DAT && TwMSG.Get                == MSG)  
				return TwOperation.PendingXfersGet;  
			if (TwDG.Control ==DG && TwDAT.PendingXfers 	 == DAT && TwMSG.Reset              == MSG)  
				return TwOperation.PendingXfersReset;  
			if (TwDG.Control ==DG && TwDAT.PendingXfers 	 == DAT && TwMSG.StopFeeder         == MSG)  
				return TwOperation.PendingXfersStopFeeder; 
			if (TwDG.Image   ==DG && TwDAT.RGBResponse 	 == DAT && TwMSG.Reset                      == MSG)  
				return TwOperation.RgbResponseReset;
			if (TwDG.Image   ==DG && TwDAT.RGBResponse 	 == DAT && TwMSG.Set                      == MSG)  
				return TwOperation.RgbResponseSet;
			if (TwDG.Control ==DG && TwDAT.SetupFileXfer 	 == DAT && TwMSG.Get                == MSG)  
				return TwOperation.SetupFileXferGet; 
			if (TwDG.Control ==DG && TwDAT.SetupFileXfer 	 == DAT && TwMSG.GetDefault         == MSG)  
				return TwOperation.SetupFileXferGetDefault;
			if (TwDG.Control ==DG && TwDAT.SetupFileXfer 	 == DAT && TwMSG.Reset              == MSG)  
				return TwOperation.SetupFileXFerReset;
			if (TwDG.Control ==DG && TwDAT.SetupFileXfer 	 == DAT && TwMSG.Set                == MSG)  
				return TwOperation.SetupFileXFerSet;
			if (TwDG.Control ==DG && TwDAT.SetupFileXfer2	 == DAT && TwMSG.Get                == MSG)  
				return TwOperation.SetupFileXfer2Get;  
			if (TwDG.Control ==DG && TwDAT.SetupFileXfer2  == DAT && TwMSG.GetDefault               == MSG)  
				return TwOperation.SetupFileXfer2GetDefault; 
			if (TwDG.Control ==DG && TwDAT.SetupFileXfer2  == DAT && TwMSG.Reset                    == MSG)  
				return TwOperation.SetupFileXfer2Reset;
			if (TwDG.Control ==DG && TwDAT.SetupFileXfer2  == DAT && TwMSG.Set                      == MSG)  
				return TwOperation.SetupFileXfer2Set;
			if (TwDG.Control ==DG && TwDAT.SetupMemXfer 	 == DAT && TwMSG.Get                == MSG)  
				return TwOperation.SetupMemXFerGet; 
			if (TwDG.Control ==DG && TwDAT.Status 		 == DAT && TwMSG.Get                == MSG)  
				return TwOperation.StatusGet;     
			if (TwDG.Control ==DG && TwDAT.UserInterface 	 == DAT && TwMSG.DisableDS          == MSG)  
				return TwOperation.UserInterfaceDisableDS; 
			if (TwDG.Control ==DG && TwDAT.UserInterface 	 == DAT && TwMSG.EnableDS           == MSG)  
				return TwOperation.UserInterfaceEnableDS;
			if (TwDG.Control ==DG && TwDAT.UserInterface 	 == DAT && TwMSG.EnableDSUIOnly     == MSG)  
				return TwOperation.UserInterfaceEnableDSUIOnly; 
			if (TwDG.Control ==DG && TwDAT.XferGroup 		 == DAT && TwMSG.Get                == MSG)  
				return TwOperation.XferGroupGet; 
			if (TwDG.Control ==DG && TwDAT.XferGroup 		 == DAT && TwMSG.Set                == MSG)  
				return TwOperation.XferGroupSet;  
			return TwOperation.NullDeviceEvent;          
		}
		
		
		public TwOperation Operation()
		{
			return TwTriplet.Operation(DG,DAT,MSG);
		}
	}
	
	public sealed class TwError:TwTriplet
	{
		internal TwStatus Status; 
		private Object target;

		internal TwError(TwDG DG,TwDAT DAT,TwMSG MSG,TwCC CC,Object target):base(DG,DAT,MSG)
		{
			this.Status=new TwStatus();
			this.Status.ConditionCode=(short)CC;
			this.target=target;
		}
		
		internal TwError(TwTriplet Tr,TwCC CC,Object target):base(Tr.DG,Tr.DAT,Tr.MSG)
		{
			this.Status=new TwStatus();
			this.Status.ConditionCode=(short)CC;
			this.target=target;
		}
		public Object GetTarget()
		{
			return this.target;
		}
		
		public TwCC Error()
		{
			return this.Status.GetTwCC();
		}

	}

	public sealed class TwProtocol
	{									// TWON_PROTOCOL...
		public const short Major	= 1;
		public const short Minor	= 9;
	}


} // namespace TwainLib

