using System;
using System.Collections;
using System.Runtime.InteropServices;
using System.Windows.Forms;


namespace PhSystems.Twain
{

	public interface ITwainCommand
	{
		void TransferReady(TwSourceManager sender);		
		void CloseRequest(TwSourceManager sender);		
		void CloseOk(TwSourceManager sender);				
		void DeviceEvent(TwSourceManager sender);
		void Failure(TwSourceManager sender);
		void Cancel(TwSourceManager sender);			
	}

	
	public interface ITwainForm:ITwainCommand
	{
		IntPtr hwndp();
	}
	

	public abstract class TwImage: IDisposable
	{
		internal IntPtr hbitmap;

		~TwImage()
		{
			this.Dispose();
		}

		
		protected TwImage(IntPtr hbitmap)
		{
			this.hbitmap= hbitmap;
		}


		public  System.Drawing.Image GetImage()
		{	
			//IntPtr punterImatge=GlobalLock(hbitmap);
			//return GetImage(GetBitmapRectangle(punterImatge),punterImatge,GetPixelInfo( punterImatge));		
			return DIBtoImage(GlobalLock(hbitmap));			
		}
		
						
		public void Dispose()
		{
			if (hbitmap!=IntPtr.Zero)
				Marshal.FreeHGlobal(hbitmap); 
		}
	
	
		private static System.Drawing.Image DIBtoImage(IntPtr dib_data) 
		{
			int									ncolors;
			int									imagesize;
			int[]								palette;
			byte[]								imagebits;
			int									bytesPerLine;
			TwBitmapInfoHeader					bmi;
			System.Drawing.Bitmap				bmp;
			System.Drawing.Imaging.BitmapData   bits;
			System.Drawing.Imaging.ColorPalette pal;
 
			bmi = (TwBitmapInfoHeader)Marshal.PtrToStructure(dib_data, typeof(TwBitmapInfoHeader)); 
			bytesPerLine = (int)((((bmi.biWidth * bmi.biBitCount) + 31) & ~31) >> 3);
			ncolors = (int)bmi.biClrUsed;
			if (ncolors == 0) 
				if (bmi.biBitCount < 24) 
					ncolors = (int)(1 << bmi.biBitCount);
			imagesize = (int)bmi.biSizeImage;
			if (imagesize == 0) 
				imagesize = bytesPerLine * bmi.biHeight;
 
			switch(bmi.biBitCount) 
			{
				case 1: 
				{       // Monochrome
					bmp = new System.Drawing.Bitmap(bmi.biWidth, bmi.biHeight, System.Drawing.Imaging.PixelFormat.Format1bppIndexed);
					bits = bmp.LockBits(new System.Drawing.Rectangle(0, 0, bmp.Width, bmp.Height), System.Drawing.Imaging.ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format1bppIndexed);
					palette = new int[2];
					break;
				}
 
				case 4: 
				{       // 4bpp
					bmp = new System.Drawing.Bitmap(bmi.biWidth, bmi.biHeight, System.Drawing.Imaging.PixelFormat.Format4bppIndexed);
					bits = bmp.LockBits(new System.Drawing.Rectangle(0, 0, bmp.Width, bmp.Height), System.Drawing.Imaging.ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format4bppIndexed);
					palette = new int[16];
					break;
				}
 
				case 8: 
				{       // 8bpp
					bmp = new System.Drawing.Bitmap(bmi.biWidth, bmi.biHeight, System.Drawing.Imaging.PixelFormat.Format8bppIndexed);
					bits = bmp.LockBits(new System.Drawing.Rectangle(0, 0, bmp.Width, bmp.Height), System.Drawing.Imaging.ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format8bppIndexed);
					palette = new int[256];
					break;
				}
 
				case 24:
				case 32: 
				{      // 32bpp
					bmp = new System.Drawing.Bitmap(bmi.biWidth, bmi.biHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
					bits = bmp.LockBits(new System.Drawing.Rectangle(0, 0, bmp.Width, bmp.Height), System.Drawing.Imaging.ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
					palette = new int[0];
					break;
				}
 
				default: 
				{
					throw new Exception("Unexpected number of bits:" + bmi.biBitCount.ToString());
				}
			}
 
			if (bmi.biBitCount < 24) 
			{
				pal = bmp.Palette;                        // Managed palette
				Marshal.Copy((IntPtr)((int)dib_data + Marshal.SizeOf(typeof(TwBitmapInfoHeader))), palette, 0, palette.Length); 
				for (int i = 0; i < ncolors; i++) 
					pal.Entries[i] = System.Drawing.Color.FromArgb(palette[i]);
			}
 
			imagebits = new byte[bytesPerLine];
			for (int y = 0; y < bmi.biHeight; y++) 
			{
				// Copy from source to managed
				Marshal.Copy((IntPtr)((int)dib_data + Marshal.SizeOf(typeof(TwBitmapInfoHeader)) + palette.Length * 4 + imagebits.Length * y), imagebits, 0, imagebits.Length);
				// Copy from managed to dest
				Marshal.Copy(imagebits, 0, (IntPtr)((int)bits.Scan0 + bits.Stride * (bmi.biHeight - 1 - y)), imagebits.Length);
			}

			/*imagebits = new byte[imagesize]; 
			// Copy from source to managed
			Marshal.Copy((IntPtr)((int)dib_data + Marshal.SizeOf(typeof(TwBitmapInfoHeader)) + palette.Length * 4), imagebits, 0, imagebits.Length);
			// Copy from managed to dest
			Marshal.Copy(imagebits, 0, bits.Scan0, imagebits.Length);*/ 

			bmp.UnlockBits(bits); 
			return bmp;
		} 

				
		private static System.Drawing.Image GetImage(System.Drawing.Rectangle bmprect, IntPtr bminfo, IntPtr pixdat)
		{
			System.Drawing.Image me = new System.Drawing.Bitmap(bmprect.Width, bmprect.Height);
			System.Drawing.Graphics g = System.Drawing.Graphics.FromImage(me);

			IntPtr hdc = g.GetHdc();
			SetDIBitsToDevice(hdc, 0, 0, bmprect.Width, bmprect.Height,0, 0, 0, bmprect.Height, pixdat, bminfo, 0);
			g.ReleaseHdc(hdc);
			return me;

		}

		private static System.Drawing.Rectangle GetBitmapRectangle( IntPtr bmpptr )
		{
			System.Drawing.Rectangle bmprect=new System.Drawing.Rectangle();
			TwBitmapInfoHeader bmi=new  TwBitmapInfoHeader();
			Marshal.PtrToStructure(bmpptr,bmi);
			bmprect.X = 0;bmprect.Y = 0;
			bmprect.Width = bmi.biWidth;
			bmprect.Height = bmi.biHeight;
			return bmprect;
		}

		private static IntPtr GetPixelInfo( IntPtr bmpptr )
		{
			System.Drawing.Rectangle bmprect=new System.Drawing.Rectangle();
			TwBitmapInfoHeader bmi=new  TwBitmapInfoHeader();
			Marshal.PtrToStructure(bmpptr,bmi);
			bmprect.X = 0;bmprect.Y = 0;
			bmprect.Width = bmi.biWidth;
			bmprect.Height = bmi.biHeight;

			if( bmi.biSizeImage == 0 )
				bmi.biSizeImage = ((((bmi.biWidth * bmi.biBitCount) + 31) & ~31) >> 3) * bmi.biHeight;

			int p = bmi.biClrUsed;
			if( (p == 0) && (bmi.biBitCount <= 8) )
				p = 1 << bmi.biBitCount;
			p = (p * 4) + bmi.biSize + (int) bmpptr;
			return (IntPtr) p;
		}

		
		[DllImport("kernel32.dll", ExactSpelling=true)]
		private static extern IntPtr GlobalLock( IntPtr handle );
		[DllImport("gdi32.dll", ExactSpelling=true)]
		private static extern int SetDIBitsToDevice( IntPtr hdc, int xdst, int ydst,int width, int height, int xsrc, int ysrc, int start, int lines,IntPtr bitsptr, IntPtr bmiptr, int color );

	}


	public sealed class TwTransferPicture:TwImage
	{
		internal TwRC rc;
		
		internal TwTransferPicture():base(IntPtr.Zero) 
		{
			this.rc=TwRC.Cancel;
		}

		public 	IntPtr Gethbitmap()
		{
			return this.hbitmap;
		}

		
	}


	public abstract class TwBase
	{

		private TwIdentity	appid;
		private TwIdentity	srcds;
		private TwError GlobalStatus;

		private bool IsEnableDS; 
		private bool IsOpenDSM; 
		private bool IsOpenDS; 
		private bool IsTranfering;

		public TwBase()
		{
			appid = new TwIdentity();
			appid.Id				= IntPtr.Zero;
			appid.Version.MajorNum	= 1;
			appid.Version.MinorNum	= 1;
			appid.Version.Language	= (short)TWCY.SPAIN;
			appid.Version.Country	= (short)TWLG.CATALAN;
			appid.Version.Info		= "LCR 1";
			appid.ProtocolMajor		= TwProtocol.Major;
			appid.ProtocolMinor		= TwProtocol.Minor;
			appid.SupportedGroups	= (int)(TwDG.Image | TwDG.Control);
			appid.Manufacturer		= "PhSystems";
			appid.ProductFamily		= "Not Freeware";
			appid.ProductName		= "PhTwain";

			srcds = new TwIdentity();
			srcds.Id = IntPtr.Zero;
		}


		private void SetGlobalStatus(TwRC rc, TwTriplet tr, Object target)
		{
			if (rc!=TwRC.Success && rc!=TwRC.XferDone)
			{
				GlobalStatus=new TwError(tr,TwCC.Success,target);
				if (DSMstatus(appid,srcds,TwDG.Control,TwDAT.Status,TwMSG.Get,GlobalStatus.Status)!=TwRC.Success)
					GlobalStatus=new TwError(tr,TwCC.OperationError,null);
			}
			if (GlobalStatus==null)
				GlobalStatus=new TwError(tr,TwCC.Success,target);
		}

		
		protected internal TwStatus GetGlobalStatus()
		{
			return GlobalStatus.Status;
		}

	
		protected internal TwRC DataSourceCapability(TwMSG MSG,TwCapability cap)
		{
			TwRC rc = DScap(appid, srcds, TwDG.Control, TwDAT.Capability, MSG, cap);
			SetGlobalStatus(rc,new TwTriplet(TwDG.Control, TwDAT.Capability, MSG),cap);
			return rc;
		}

		protected internal TwRC DataSourceManagerParent(TwMSG MSG,IntPtr hwndp)
		{
			TwRC rc=TwRC.Success;
			if( !(TwMSG.CloseDSM==MSG && appid.Id == IntPtr.Zero))
				rc= DSMparent(appid, IntPtr.Zero, TwDG.Control, TwDAT.Parent, MSG, ref hwndp );
			if (TwMSG.CloseDSM==MSG)
				appid.Id = IntPtr.Zero;
			if (MSG==TwMSG.OpenDSM && rc==TwRC.Success)
				IsOpenDSM=true;
			if (MSG==TwMSG.CloseDSM && rc==TwRC.Success)
				IsOpenDSM=false;
			SetGlobalStatus(rc,new TwTriplet(TwDG.Control, TwDAT.Parent, MSG),null);
			return rc;		
		}

		protected internal TwRC DataSourceManagerUserInfo(TwMSG MSG,IntPtr hwnd)
		{
			TwUserInterface	guif = new TwUserInterface();
			if (MSG!=TwMSG.DisableDS)
				guif.Initilalize(hwnd); 
			TwRC rc = DSuserif( appid, srcds, TwDG.Control, TwDAT.UserInterface,MSG, guif );
			if (MSG==TwMSG.EnableDS && rc==TwRC.Success)
				IsEnableDS=true;
			if (MSG==TwMSG.DisableDS && rc==TwRC.Success)
				IsEnableDS=false;
			SetGlobalStatus(rc,new TwTriplet(TwDG.Control, TwDAT.UserInterface,MSG),null);
			return rc;
		}

		protected internal TwRC DataSourceManagerIdentity(TwMSG MSG)
		{
			try
			{
				TwRC rc=DSMident(appid, IntPtr.Zero, TwDG.Control, TwDAT.Identity, MSG, srcds);
				if (MSG==TwMSG.OpenDS && rc==TwRC.Success)
					IsOpenDS=true;
				if (MSG==TwMSG.CloseDS && rc==TwRC.Success)
					IsOpenDS=false;
				SetGlobalStatus(rc,new TwTriplet(TwDG.Control, TwDAT.Identity, MSG),null);
				return rc;
			}
			catch
			{
			}
			return TwRC.Failure;
		}

		protected internal TwRC DataSourceManagerImageInfo(TwMSG operation, TwImageInfo ImageInfo)
		{
			TwRC rc = DSiinf( appid, srcds, TwDG.Image, TwDAT.ImageInfo, operation, ImageInfo );
			SetGlobalStatus(rc,new TwTriplet(TwDG.Image, TwDAT.ImageInfo, operation),ImageInfo);
			return rc;
		}

		protected internal TwRC DataSourceManagerImageTransfer(ref IntPtr hbitmap)
		{
			TwRC rc = DSixfer( appid, srcds, TwDG.Image, TwDAT.ImageNativeXfer, TwMSG.Get, ref hbitmap );
			SetGlobalStatus(rc,new TwTriplet(TwDG.Image, TwDAT.ImageNativeXfer, TwMSG.Get),null);
			return rc;

		}
		protected internal TwRC DataSourceManagerPendingInfo(TwMSG MSG,ref TwPendingXfers pxfr)
		{
			TwRC rc = DSpxfer( appid, srcds, TwDG.Control, TwDAT.PendingXfers, MSG, pxfr );
			if(MSG==TwMSG.Reset || MSG==TwMSG.EndXfer)
				this.IsTranfering= (MSG==TwMSG.EndXfer && pxfr.Count!=0 && (rc==TwRC.Success || rc==TwRC.XferDone));
//				this.IsTranfering=!(rc==TwRC.Success || rc==TwRC.XferDone)&& pxfr.Count!=0;
			SetGlobalStatus(rc,new TwTriplet(TwDG.Control, TwDAT.PendingXfers, MSG),pxfr);
			return rc;
		}

		protected internal TwRC DataSourceManagerEvent(ref TwEvent evtmsg)
		{
			evtmsg.Message = 0;
			return DSevent(appid, srcds, TwDG.Control, TwDAT.Event, TwMSG.ProcessEvent, ref evtmsg);
		}

		
		protected internal bool IsTWTranfering()
		{
			return IsTranfering;
		}

		protected internal bool IsTWEnablesDS()
		{
			return IsEnableDS;
		}
		protected internal bool IsTWOpenedDSM()
		{
			return IsOpenDSM;
		}
		protected internal bool IsTWOpenedDS()
		{
			return IsOpenDS;
		}

		
		protected internal bool IsApplicationAvailable()
		{
			return appid.Id != IntPtr.Zero;
		}

		protected internal bool IsDataSourceAvailable()
		{
			return srcds.Id != IntPtr.Zero;
		}


		internal TwainCommand PassMessage(TwRC rc,short Message)
		{
			if( Message == (short) TwMSG.XFerReady)
			{
				this.IsTranfering=true;
				return TwainCommand.TransferReady;
			}
			if( Message == (short) TwMSG.CloseDSReq)
				return TwainCommand.CloseRequest;
			if( rc==TwRC.Cancel )
				return TwainCommand.Cancel;
			if( rc==TwRC.Failure)
				return TwainCommand.Failure;
			if( Message == (short) TwMSG.CloseDSOK || Message == (short) TwMSG.CloseDS)
				return TwainCommand.CloseOk;
			if( Message == (short) TwMSG.DeviceEvent )
				return TwainCommand.DeviceEvent;
			if( rc == TwRC.NotDSEvent )
				return TwainCommand.Not;
			return TwainCommand.Null;
		}	

		
		public static int ScreenBitDepth 
		{
			get 
			{
				IntPtr screenDC = CreateDC( "DISPLAY", null, null, IntPtr.Zero );
				int bitDepth = GetDeviceCaps( screenDC, 12 );
				bitDepth *= GetDeviceCaps( screenDC, 14 );
				DeleteDC( screenDC );
				return bitDepth;
			}
		}

		public static bool CheckForTwain() 
		{    
			string sbWinDir = "";

			// 1. Look in WIN.INI for "[TWAIN]"
			bool bTWAINFound = false;
			if (System.IO.File.Exists(System.IO.Path.Combine(sbWinDir.ToString().Trim(), "win.ini"))) 
			{
				System.IO.StreamReader reader = System.IO.File.OpenText(System.IO.Path.Combine(sbWinDir.ToString().Trim(), "win.ini"));
				String strLine = reader.ReadLine();
				while (strLine != null &&  !bTWAINFound)
				{    
					bTWAINFound = (strLine.IndexOf("[TWAIN]", 0, strLine.Length) >= 0);
					if (!bTWAINFound)
					{
						strLine = reader.ReadLine();
					}
					// read the next line
				}
				reader.Close();
			}
			// 2. Look for TWAIN_32.DLL
			return (bTWAINFound || System.IO.File.Exists(System.IO.Path.Combine(System.IO.Directory.GetParent(System.Reflection.Assembly.GetExecutingAssembly().Location).ToString(), "TWAIN_32.DLL")) || System.IO.File.Exists(System.IO.Path.Combine(sbWinDir.ToString().Trim(), "TWAIN_32.DLL"))|| System.IO.File.Exists(System.IO.Path.Combine(System.Environment.SystemDirectory, "TWAIN_32.DLL")));
		}

		public TwIdentity GetIdentity()
		{
			return srcds;
		}
		public TwError LastError()
		{
			return this.GlobalStatus;
		}


		#region DSM entry point

		// ------ DSM entry point DAT_ variants:
		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DSMparent( [In, Out] TwIdentity origin, IntPtr zeroptr, TwDG dg, TwDAT dat, TwMSG msg, ref IntPtr refptr );

		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DSMident( [In, Out] TwIdentity origin, IntPtr zeroptr, TwDG dg, TwDAT dat, TwMSG msg, [In, Out] TwIdentity idds );

		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DSMstatus( [In, Out] TwIdentity origin, [In, Out] TwIdentity dest, TwDG dg, TwDAT dat, TwMSG msg, [In, Out] TwStatus dsmstat );

		// ------ DSM entry point DAT_ variants to DS:
		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DSuserif( [In, Out] TwIdentity origin, [In, Out] TwIdentity dest, TwDG dg, TwDAT dat, TwMSG msg, TwUserInterface guif );

		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DSevent( [In, Out] TwIdentity origin, [In, Out] TwIdentity dest, TwDG dg, TwDAT dat, TwMSG msg, ref TwEvent evt );

		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DSstatus( [In, Out] TwIdentity origin, [In] TwIdentity dest, TwDG dg, TwDAT dat, TwMSG msg, [In, Out] TwStatus dsmstat );

		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DScap( [In, Out] TwIdentity origin, [In] TwIdentity dest, TwDG dg, TwDAT dat, TwMSG msg, [In, Out] TwCapability capa );

		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DSiinf( [In, Out] TwIdentity origin, [In] TwIdentity dest, TwDG dg, TwDAT dat, TwMSG msg, [In, Out] TwImageInfo imginf );

		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DSixfer( [In, Out] TwIdentity origin, [In] TwIdentity dest, TwDG dg, TwDAT dat, TwMSG msg, ref IntPtr hbitmap );

		[DllImport("twain_32.dll", EntryPoint="#1")]
		private static extern TwRC DSpxfer( [In, Out] TwIdentity origin, [In] TwIdentity dest, TwDG dg, TwDAT dat, TwMSG msg, [In, Out] TwPendingXfers pxfr );

		[DllImport("gdi32.dll", ExactSpelling=true)]
		private static extern int GetDeviceCaps( IntPtr hDC, int nIndex );

		[DllImport("gdi32.dll", CharSet=CharSet.Auto)]
		private static extern IntPtr CreateDC( string szdriver, string szdevice, string szoutput, IntPtr devmode );

		[DllImport("gdi32.dll", ExactSpelling=true)]
		private static extern bool DeleteDC( IntPtr hdc );

		#endregion

	}


	public abstract class TwBaseManager:TwBase,IMessageFilter
	{

		private WINMSG		winmsg;
		private TwEvent		evtmsg;
		private bool	 isDisable;
		private System.Collections.ArrayList Capabilitys;

		protected abstract  void Finish();	
		protected abstract  IntPtr Gethwnd();
		protected abstract bool IsPendingXfers();
		//protected abstract TwPendingXfers PendingXfers();
		protected abstract bool FilterMessage(TwainCommand cmd);

		protected abstract  bool TWTransferImage(ref IntPtr Image);
		protected abstract  bool TWOpenTransferImage();
		protected abstract  bool TWCloseTransferImage(bool IsCancel);
		protected abstract  bool TWOpenDSM();
		protected abstract  bool TWCloseDSM();
		protected abstract  bool TWOpenDS();
		protected abstract  bool TWCloseDS();
		protected abstract  bool TWDisableDS();
		protected abstract  bool TWGetIdentity();
		protected abstract  bool TWEnableDS(bool ShowUI);
		protected abstract  bool TWSetCapabilitis(TwCapability[] Capabilitis);

		public class ITwTransferPictureEnumerator:IEnumerator
		{
			private TwBaseManager Host;
			private TwTransferPicture image;
			internal ITwTransferPictureEnumerator(TwBaseManager Host)
			{
				this.Host=Host;
				this.Host.TWOpenTransferImage();
			}

		
			#region Miembros de IEnumerator

			public void Reset()
			{
				image=null;
				this.Host.TWCloseTransferImage(true);
			}

			public object Current
			{
				get
				{
					return image;
				}
			}

			public bool MoveNext()
			{
				image=null;
				if (this.Host.IsPendingXfers())
					image= (TwTransferPicture)this.TransferPicture();
				if (image!=null)
					return (image.rc==TwRC.Success);
				return false;
			}

			#endregion

			private TwImage TransferPicture()
			{
				TwTransferPicture rc=new TwTransferPicture();
				TwImageInfo	iinf = new TwImageInfo();
				rc.rc=TwRC.Failure;
				if (this.Host.IsPendingXfers())
				{			
					rc.rc=this.Host.DataSourceManagerImageInfo(TwMSG.Get,iinf);
					if(this.Host.onfilterMessage(rc.rc,TwMSG.Null))
						if (!this.Host.TWTransferImage(ref rc.hbitmap))
							rc.rc=TwRC.Failure;
					if (!this.Host.TWCloseTransferImage(!(rc.rc==TwRC.Success || rc.rc==TwRC.XferDone)))
						rc.rc=TwRC.Failure;				
					return rc;
				}
				this.Host.onfilterMessage(rc.rc,TwMSG.Null);
				return rc;
			}


			public TwImage CurrentPicture()
			{
				return (TwImage)this.Current;
			}
			
			public TwainCommand Error()
			{
				if (image!=null)
					return ((TwBase)this.Host).PassMessage(image.rc,(short)TwMSG.Null); 	
				return ((TwBase)this.Host).PassMessage((TwRC.Failure),(short)TwMSG.Null);
			}


		}


		protected TwBaseManager()
		{
			evtmsg.EventPtr = Marshal.AllocHGlobal(Marshal.SizeOf(winmsg));
			Capabilitys=new System.Collections.ArrayList();
			Capabilitys.Add(new TwCapability(TwCap.XFERCOUNT, (short)-1 ));
		}

		~TwBaseManager()
		{
			this.Finish();
			if (evtmsg.EventPtr!=IntPtr.Zero)
				Marshal.FreeHGlobal(evtmsg.EventPtr);
			if (Capabilitys!=null) 
			{
				System.Collections.IEnumerator enumerator=Capabilitys.GetEnumerator();
				while (enumerator.MoveNext())
					((TwCapability)enumerator.Current).Dispose();
				this.Capabilitys.Clear();	
			}

		}


		bool IMessageFilter.PreFilterMessage(ref Message m )
		{		
			return this.FilterMessage(this.MessageToTwCommand(ref m));
		}

		
		private TwainCommand MessageToTwCommand(ref Message m)
		{

			int pos = GetMessagePos();

			winmsg.hwnd		= m.HWnd;
			winmsg.message	= m.Msg;
			winmsg.wParam	= m.WParam;
			winmsg.lParam	= m.LParam;
			winmsg.time		= GetMessageTime();
			winmsg.x		= (short) pos;
			winmsg.y		= (short) (pos >> 16);
			
			Marshal.StructureToPtr(winmsg,evtmsg.EventPtr, false);
			TwRC rc=this.DataSourceManagerEvent(ref evtmsg);
			return ((TwBase)this).PassMessage(rc,evtmsg.Message);

		}
		

		public bool IsAvailable()
		{
			if (isDisable)
				return false;
			if (this.IsTWEnablesDS())
				if (!this.TWDisableDS())
					return onfilterMessage(TwRC.Failure,TwMSG.Null);
			if (this.IsTWOpenedDS())
				if(!this.TWCloseDS())
					return onfilterMessage(TwRC.Failure,TwMSG.Null);
			if (!this.IsTWOpenedDSM())
				if (InitApplication())
					return onfilterMessage(TwRC.Failure,TwMSG.Null);
			return true;
		}


		protected void Disable()
		{
			isDisable=true;
		}

		protected bool onfilterMessage(TwRC rc,TwMSG MSG)
		{
			//if (rc==TwRC.Success || rc==TwRC.XferDone)
			//	return true;
			return this.FilterMessage(((TwBase)this).PassMessage(rc,(short)MSG)); 					
		}

		protected bool InitApplication()
		{						
			if (this.IsTWEnablesDS())
				if (!this.TWDisableDS())
					return false;
			if (this.IsTWOpenedDS())
				if(!this.TWCloseDS())
					return false;
			if (this.IsTWOpenedDSM())
				if(!this.TWCloseDSM())
					return false;
			if (!this.TWOpenDSM())
				return false;
			else 
				if (!this.TWGetIdentity())
					this.TWCloseDSM();
			return this.IsTWOpenedDSM();
		}

		
		public void Select()
		{
			if (this.IsAvailable())
				this.onfilterMessage(this.DataSourceManagerIdentity(TwMSG.UserSelect),PhSystems.Twain.TwMSG.UserSelect);		
		}
		
		public bool AddCapability(TwCap cap, short sval)
		{
			return this.AddCapability(new TwCapability(cap,sval));
		}
		public bool AddCapability(TwCapability cap)
		{
			System.Collections.ArrayList delete=new System.Collections.ArrayList();

			IEnumerator Enumerator=Capabilitys.GetEnumerator();
			while (Enumerator.MoveNext())
				if (((TwCapability)Enumerator.Current).Cap==cap.Cap)
					delete.Add(Enumerator.Current);
					
			Enumerator=delete.GetEnumerator();
			while (Enumerator.MoveNext())
			  Capabilitys.Remove(Enumerator.Current); 	

			//if (!this.AcceptCapability((TwCap)cap.Cap))
			//	return false;
			Capabilitys.Add(cap);
			return true;
		}

		public void Acquire(bool ShowUI)
		{
			if (this.IsAvailable())
				if (this.TWOpenDS())
					if (this.TWSetCapabilitis((TwCapability[])Capabilitys.ToArray(typeof(TwCapability))))		
						this.TWEnableDS(ShowUI);	
					else
						this.TWCloseDS();
		}

		public bool AcceptCapability(TwCapability cap)
		{
			bool ret=false;
			if (!this.IsAvailable())
				return ret;
			bool open=!this.IsTWOpenedDS();				
			if (this.TWOpenDS())
			{					
				ret=this.onfilterMessage(this.DataSourceCapability(TwMSG.QuerySupport, cap),PhSystems.Twain.TwMSG.QuerySupport);
				if (open)		
					this.TWCloseDS();
			}
			return ret;
		}

		public bool AcceptCapability(TwCap Capability)
		{
			return this.AcceptCapability(new TwCapability(Capability));
		}
		
		public TwCapability GetCapability(TwCap cap)
		{
			TwCapability tcap=new TwCapability(cap);
			bool ret=false;
			if (!this.IsAvailable())
				return null;
			bool open=!this.IsTWOpenedDS();				
			if (this.TWOpenDS())
			{
				ret=this.onfilterMessage(this.DataSourceCapability(TwMSG.Get, tcap),PhSystems.Twain.TwMSG.Get);
				if (open)		
					this.TWCloseDS();					
			}
			if (ret)
				return tcap;
			return null;
		}

		public TwCap[] SupportedCapabilitis()
		{
			ArrayList llista =new ArrayList();
			//this.GetCapability(TwCap.SUPPORTEDCAPS).GetTwArray();
			//this.GetCapability(TwCap.IBITDEPTH).GetTwEnumeration();
			Twain.TwContainerCapability Capability= this.GetCapability(TwCap.SUPPORTEDCAPS).GetContainerCapabilityValue();
			if (Capability == null) 
				return new TwCap[]{};
			if (!(Capability is TwArray)) 
				return new TwCap[]{};
			if ((((TwArray)Capability).ItemList)==null)
				return new TwCap[]{};
			for(int i=0;i<((TwArray)Capability).ItemList.Length;i++)
				llista.Add((TwCap)((TwArray)Capability).ItemList[i]);
			return (TwCap[])llista.ToArray(typeof(TwCap));			
		}

		public ITwTransferPictureEnumerator TransferPictures()
		{
			return (new ITwTransferPictureEnumerator(this));
		}


		#region DllImport

		[DllImport("kernel32.dll", ExactSpelling=true)]
		internal static extern IntPtr GlobalAlloc( int flags, int size );
		[DllImport("kernel32.dll", ExactSpelling=true)]
		internal static extern IntPtr GlobalLock( IntPtr handle );
		[DllImport("kernel32.dll", ExactSpelling=true)]
		internal static extern bool GlobalUnlock( IntPtr handle );
		[DllImport("kernel32.dll", ExactSpelling=true)]
		internal static extern IntPtr GlobalFree( IntPtr handle );

		[DllImport("user32.dll", ExactSpelling=true)]
		private static extern int GetMessagePos();
		[DllImport("user32.dll", ExactSpelling=true)]
		private static extern int GetMessageTime();

		[StructLayout(LayoutKind.Sequential, Pack=4)]
		internal struct WINMSG
		{
			public IntPtr		hwnd;
			public int			message;
			public IntPtr		wParam;
			public IntPtr		lParam;
			public int			time;
			public int			x;
			public int			y;
		}


		#endregion

	}


	public abstract class TwBaseSource:TwBaseManager
	{

		protected enum TwOperationMessage
		{
			DataSource=0,
			DataSourceManager=1,
			DataSourceInterface=2,
			DataSourceTransfering=3
		}
		protected enum TwOperationType
		{
			Begining=0,
			Ending=1
		}
		
			
		private TwPendingXfers pxfr = new TwPendingXfers();

		protected abstract void state(TwOperationType type, TwOperationMessage Operation);

		protected override bool TWTransferImage(ref IntPtr Image)
		{
			if (this.IsTWTranfering())
				if( this.onfilterMessage(DataSourceManagerImageTransfer(ref Image),PhSystems.Twain.TwMSG.Get))
					return true;
			return false;
		}

		protected override bool TWDisableDS()
		{
			if (!this.IsTWEnablesDS())
				return true;
			if (this.onfilterMessage(this.DataSourceManagerUserInfo(TwMSG.DisableDS,IntPtr.Zero),PhSystems.Twain.TwMSG.DisableDS))
				this.state(TwOperationType.Ending,TwOperationMessage.DataSourceInterface);
			return !this.IsTWEnablesDS();
		}
		
		protected override bool TWEnableDS(bool ShowUI)
		{
			if (this.IsTWEnablesDS())
				return true;
			if (this.onfilterMessage(DataSourceManagerUserInfo(TwMSG.EnableDS,ShowUI?this.Gethwnd():IntPtr.Zero),PhSystems.Twain.TwMSG.EnableDS))
				this.state(TwOperationType.Begining,TwOperationMessage.DataSourceInterface);
			return this.IsTWEnablesDS();
		}
	
		protected override bool TWOpenDSM()
		{
			if (this.IsTWOpenedDSM())
				return true;
			if (this.onfilterMessage(this.DataSourceManagerParent(TwMSG.OpenDSM,this.Gethwnd() ),PhSystems.Twain.TwMSG.OpenDSM))
				this.state(TwOperationType.Begining,TwOperationMessage.DataSourceManager);
			return this.IsTWOpenedDSM();
		}
		protected override bool TWCloseDSM()
		{
			if (this.IsTWOpenedDSM())
				return true;
			if (this.onfilterMessage(this.DataSourceManagerParent(TwMSG.CloseDSM,this.Gethwnd()),PhSystems.Twain.TwMSG.CloseDSM))
				this.state(TwOperationType.Ending,TwOperationMessage.DataSourceManager);
			return !this.IsTWOpenedDSM();
		}

		protected override bool TWOpenDS()
		{
			if (this.IsTWOpenedDS())
				return true;
			if (this.onfilterMessage(this.DataSourceManagerIdentity(TwMSG.OpenDS),PhSystems.Twain.TwMSG.OpenDS))
				this.state(TwOperationType.Begining,TwOperationMessage.DataSource);
			return this.IsTWOpenedDS();
		}
	
		protected override bool TWCloseDS()
		{
			if (!this.IsTWOpenedDS())
				return true;
			if (this.onfilterMessage(this.DataSourceManagerIdentity(TwMSG.CloseDS),PhSystems.Twain.TwMSG.CloseDS))
				this.state(TwOperationType.Ending,TwOperationMessage.DataSource);
			return !this.IsTWOpenedDS();
		}

		protected override bool TWGetIdentity()
		{
			return this.onfilterMessage(this.DataSourceManagerIdentity(TwMSG.GetDefault),PhSystems.Twain.TwMSG.GetDefault);
		}

		protected override bool TWSetCapabilitis(TwCapability[] Capabilitis)
		{
			if (Capabilitis==null)
				return false;
			System.Collections.IEnumerator enumerator=Capabilitis.GetEnumerator();
			while (enumerator.MoveNext())
				if(!this.onfilterMessage(this.DataSourceCapability(TwMSG.Set,(TwCapability)enumerator.Current),PhSystems.Twain.TwMSG.Set))
					return false;
			return true;
		}
	
		protected override bool TWCloseTransferImage(bool IsCancel)
		{
			if (this.IsTWTranfering())
				if( this.onfilterMessage(DataSourceManagerPendingInfo(IsCancel?TwMSG.Reset:TwMSG.EndXfer,ref this.pxfr),PhSystems.Twain.TwMSG.EndXfer))
					return true;
			return false;
		}

		protected override bool TWOpenTransferImage()
		{
			if (pxfr==null)
				return false;
			if (this.onfilterMessage(this.IsTWTranfering()?TwRC.Success:TwRC.Failure,TwMSG.Null))
				pxfr.Count=1;
			else
				pxfr.Count=0;
			return pxfr.Count==1;
		}
		
		/*
		protected override TwPendingXfers PendingXfers()
		{
			return pxfr;
		}
		*/
		protected override bool IsPendingXfers()
		{
			if (pxfr==null)
				return false;
			return pxfr.Count!=0;
		}

		protected override void Finish()
		{
			if (this.IsTWEnablesDS())
				if (!this.TWDisableDS())
					return;
			if (this.IsTWOpenedDS())
				this.TWCloseDS();
			else
				return;
			if (this.IsTWOpenedDSM())
				this.TWCloseDSM();
			else
				return;
		}


	}


	public class TwSourceManager:TwBaseSource
	{
		private bool msgfilter=false;
		private IntPtr hwnd=IntPtr.Zero;
		private PostFilterMessage myPreFilterMessage=null;

		public delegate void PostFilterMessage(TwainCommand TC);

		protected TwSourceManager()
		{
		}

		~TwSourceManager()
		{
			this.myPreFilterMessage=null;
			this.hwnd=IntPtr.Zero;
		}

		
		public TwSourceManager(PostFilterMessage myFMessage,IntPtr hwnd)
		{
			this.InitApplication(myFMessage,hwnd);
		}


		protected override IntPtr Gethwnd()
		{
			return this.hwnd;
		}

		protected override bool FilterMessage(TwainCommand cmd)
		{
			bool ret=false;
			switch( cmd )
			{
				default:
				{
					return ret;
				}
				case TwainCommand.Null:
				{
					return (!ret);
				}
				case TwainCommand.Not:
				{
					return ret;
				}
				case TwainCommand.CloseRequest:
				{
					//if (this.myPreFilterMessage!=null)
					//	this.myPreFilterMessage(cmd); 
					if (this.TWDisableDS())
						ret=this.TWCloseDS();
					if (this.myPreFilterMessage!=null)
						this.myPreFilterMessage(cmd); 
					return ret;
				}
				case TwainCommand.CloseOk:
				{
					//if (this.myPreFilterMessage!=null)
					//	this.myPreFilterMessage(cmd); 				
					if (this.TWDisableDS())
						ret=this.TWCloseDS();
					if (this.myPreFilterMessage!=null)
						this.myPreFilterMessage(cmd); 				
					return ret;
				}
				case TwainCommand.DeviceEvent:
				{
					if (this.myPreFilterMessage!=null)
						this.myPreFilterMessage(cmd); 				
					return (!ret);
				}
				case TwainCommand.TransferReady:
				{
					if (this.myPreFilterMessage!=null)
						this.myPreFilterMessage(cmd);
					else
						this.FilterMessage(TwainCommand.Failure);
					if (this.IsTWTranfering())
						this.TWCloseTransferImage(true);
					if (this.TWDisableDS())
						ret=this.TWCloseDS();
					return ret;
				}
				case TwainCommand.Failure:
				{
					if (this.myPreFilterMessage!=null)
						this.myPreFilterMessage(cmd);				
					return (ret);
				}
				case TwainCommand.Cancel:
				{
					if (this.myPreFilterMessage!=null)
						this.myPreFilterMessage(cmd);				
					return (ret);
				}
			}
		}
			
		protected override void state(TwOperationType type, TwOperationMessage Operation)
		{
			if (Operation==TwOperationMessage.DataSourceInterface)
				OperationInterface(type);
		}
		
	
		protected bool InitApplication(PostFilterMessage mfmsg,IntPtr hwnd)
		{
			this.hwnd=hwnd;
			this.myPreFilterMessage=mfmsg;
			if (!(this.hwnd==IntPtr.Zero || this.myPreFilterMessage==null))
				return base.InitApplication();
			base.Disable();
			return false;
		}


		private  void OperationInterface(TwOperationType type)
		{
			if(msgfilter && type==TwOperationType.Ending)
				Application.RemoveMessageFilter(this);
			if(!msgfilter && type==TwOperationType.Begining)
				Application.AddMessageFilter(this);
			msgfilter=!(msgfilter && type==TwOperationType.Ending)||(!msgfilter && type==TwOperationType.Begining);
		}
		

	}


	public sealed class TwFormHelper:TwSourceManager
	{
		private bool Available;

		private ITwainForm Form;

		public TwFormHelper(ITwainForm Form)
		{			
			this.Form=Form;
			Available=base.InitApplication(new TwSourceManager.PostFilterMessage(this.TwainCommandFilter),Form.hwndp());
		}

		~TwFormHelper()
		{
			this.Form=null;
			if (this.IsAvailable())
				base.Finish();
		}
				
		
		public new bool  IsAvailable()
		{
			if (base.IsAvailable())
				return (Available);
			return false;
		}

	
		private void TwainCommandFilter(TwainCommand TC)
		{
			if (this.Form==null)
				return;
			switch (TC)
			{
				case TwainCommand.CloseOk:
				{
					this.Form.CloseOk(this); 
					break;
				}
				case TwainCommand.CloseRequest:
				{
					this.Form.CloseRequest(this); 
					break;
				}
				case TwainCommand.TransferReady:
				{
					this.Form.TransferReady(this); 
					break;
				}
				case TwainCommand.DeviceEvent:
				{
					this.Form.DeviceEvent(this); 
					break;
				}
				case TwainCommand.Cancel:
				{
					this.Form.Cancel(this); 
					break;
				}
				case TwainCommand.Failure:
				{
					this.Form.Failure(this); 
					break;
				}
			}
		}
	}


}

